/*****************************************************************************************************************************************
October 21 2017: add the following feature by Huaichun Wang

Use -aafreq to let the program uses a use-defined aa frequency file (i.e., +F for a protein model).
Note: the amino acids in the frequency file is in the order:
ala, arg, asn, asp, cys, gln, glu, gly, his, ileu, leu, lys, met, phe, pro, ser, thr, trp, tyr, val

For C20 and C60 models, use -wt my_weight_file so that the simulated lengths for the F+C20 pr F+C60 classes will be proportional 
to their weights in the weight file.

For C20 there are 21 weights, in the order F, C1, C2, ..., C20.
./SiteSpecific.seq-gen -mLG -g 4 -a 0.75 -p 21 -C20 2 -l 1000 example.tre -wt example.c20.wt.dat

For C60 there are 61 weights, in the order F, C1, C2, ..., C60.
./SiteSpecific.seq-gen -mLG -g 4 -a 0.75 -p 61 -C60 2 -l 1000 example.tre -wt example.c60.wt.dat

Note: under -wt, -C20 or -C60 can set any number greater than 1, as the weights and the lengths (-l) will determine each profile's length.

24 Feb 2006: added general covarion model by Matt Spencer
   24 Oct 2016: added site-specific frequency profiles: C20, C60 and par-site partition (SSF) by Huaichun Wang
   please consult README.txt for usage information.

   Sequence Generator - seq-gen, version 1.3.2
   Copyright (c)1996-2004, Andrew Rambaut & Nick Grassly
   Department of Zoology, University of Oxford			
   All rights reserved.                          

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

     1. Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.

     2. Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.

     3. The names of its contributors may not be used to endorse or promote 
        products derived from this software without specific prior written 
        permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


   Any feedback is very welcome.
   http://evolve.zoo.ox.ac.uk/software/Seq-Gen/
   email: andrew.rambaut@zoo.ox.ac.uk
****************************************************************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>

#include "global.h"
#include "treefile.h"
#include "evolve.h"
#include "model.h"
#include "nucmodels.h"
#include "aamodels.h"
#include "progress.h"
#include "twister.h"
#include "covarion.h"

#define PROGRAM_NAME "seq-gen-aminocov"
#define VERSION_NUMBER "Version 1.3, adapted from seq-gen 1.3.2"

int treeFile, textFile, numDatasets, numTrees;
int scaleTrees, scaleBranches, ancestorSeq, writeAncestors, writeRates;
int writeLengths, writeHidStates; /*added from Ane code*/
int *partitionLengths;
double *partitionRates;
double treeScale, branchScale;
double covarionSpeed;/*added from Ane code*/
char treeFileName[256];
char textFileName[256];
char site_aafreqFileName[256]; /* user-define per-site aa freq name */
char CFweightFileName[256];
char aaFreqFileName[256];    /* user-define aa freq name */

int hasAlignment, numSequences, numAlignmentSites;
char **names;
char **sequences;

FILE *tree_fv;

int simuPartitionLen;
int C20, C60, SSF, havingWeight, user_defined_aafreq;


/* prototypes */

static void PrintTitle();
static void PrintUsage();
static void PrintVerbose(FILE *fv);
static void ReadParams();
static void ReadFileParams();
static void AllocateMemory();
static void ReadFile();
static int OpenTreeFile();
 

/* functions */

static void PrintTitle()
{
	fprintf(stderr, "Sequence Generator - %s\n", PROGRAM_NAME);
	fprintf(stderr, "%s\n", VERSION_NUMBER);
	fprintf(stderr, "(c) Copyright, 1996-2004 Andrew Rambaut and Nick Grassly\n");
	fprintf(stderr, "Department of Zoology, University of Oxford\n");
	fprintf(stderr, "South Parks Road, Oxford OX1 3PS, U.K.\n\n");
}

static void PrintUsage()
{ 
	fprintf(stderr, "Usage: seq-gen-aminocov [-m MODEL] [-l #] [-n #] [-p #] [-s # | -d #] [-k #]\n");
/*
"[-i #]" inserted in the following line

7 Oct 2011 -- Lars Jermiin (lars.jermiin@csiro.au)
*/
	fprintf(stderr, "               [-c #1 #2 #3 | -a # [-g #]] [-i #] [-f e | #] [-t # | -r #]\n");
	fprintf(stderr, "               [-z #] [-o[p][r][n]] [-w[a][r][h][l]] [-x NAME] [-q] [-h]\n");
	fprintf(stderr, "                   [-v #1 #2] [-u #1 #2] [-y #1 #2 #3 #4] [treefile]\n");	/* -v added from Ane code*/
	fprintf(stderr, "  -l: # = sequence length [default = 1000].\n");
	fprintf(stderr, "  -n: # = simulated datasets per tree [default = 1].\n");
	fprintf(stderr, "  -p: # = number of partitions (and trees) per sequence [default = 1].\n");
	fprintf(stderr, "  -s: # = branch length scaling factor [default = 1.0].\n");
	fprintf(stderr, "  -d: # = total tree scale [default = use branch lengths].\n");
	fprintf(stderr, "  -k: # = use sequence k as ancestral (needs alignment) [default = random].\n");

/*
"CPREV45, MTART," inserted in the following line

7 Oct 2011 -- Lars Jermiin (lars.jermiin@csiro.au)
*/
	fprintf(stderr, "  -m: MODEL = HKY, F84, GTR, JTT, WAG, LG, PAM, BLOSUM, MTREV, CPREV45, MTART, GENERAL\n");
	fprintf(stderr, "      HKY, F84 & GTR are for nucleotides the rest are for amino acids\n");
	fprintf(stderr, "  -a: # = shape (alpha) for gamma rate heterogeneity [default = none].\n");
	fprintf(stderr, "  -g: # = number of gamma rate categories [default = continuous].\n");
	fprintf(stderr, "  -i: # = proportion of invariable sites [default = 0.0].\n");

	fprintf(stderr, "\n Nucleotide model specific options:\n");
	fprintf(stderr, "  -c: #1 #2 #3 = rates for codon position heterogeneity [default = none].\n");
	fprintf(stderr, "  -t: # = transition-transversion ratio [default = equal rate].\n");
	fprintf(stderr, "  -r: #1 #2 #3 #4 #5 #6= general rate matrix [default = all 1.0].\n");
	fprintf(stderr, "  -f: #A #C #G #T = nucleotide frequencies [default = all equal].\n");
	
	fprintf(stderr, "\n Amino Acid model specific options:\n");
	fprintf(stderr, "      specify using the order ARNDCQEGHILKMFPSTWYV\n");
	fprintf(stderr, "  -r: #1 .. #190 = general rate matrix [default = all 1.0].\n");
	fprintf(stderr, "  -f: #1 .. #20 = amino acid frequencies e=equal [default = matrix freqs].\n");
	fprintf(stderr, "  -aafreq: user defined amino acid frequency file; the order of the amino acids should be: ala, arg, asn, asp, cys, gln, glu, gly, his, ileu, leu, lys, met, phe, pro, ser, thr, trp, tyr, val. [default = none: the amino acid frequencies of the proetin model will be used]. \n");
	
	fprintf(stderr, "\n Miscellaneous options:\n");
	fprintf(stderr, "  -z: # = seed for random number generator [default = system generated].\n");
	fprintf(stderr, "  -o: Output file format [default = PHYLIP]\n");
	fprintf(stderr, "    p PHYLIP format\n");
	fprintf(stderr, "    r relaxed PHYLIP format\n");
	fprintf(stderr, "    n NEXUS format\n");
	fprintf(stderr, "  -w: Write additional information [default = none]\n");
	fprintf(stderr, "    a Write ancestral sequences for each node\n");
	fprintf(stderr, "    r Write rate for each site\n");
	fprintf(stderr, "    h Write covarion state (on = + off = - in Tuffley & Steel/Huelsenbeck,\n");
	fprintf(stderr, "    1...g for Galtier covarion sites, -1...-g for Galtier RAS sites),\n");
	fprintf(stderr, "    for each site and each node.\n");
	fprintf(stderr, "    l Write effective length for each site (covarion process)\n");

	fprintf(stderr, "  -x: NAME = a text file to insert after every dataset [default = none].\n");
	fprintf(stderr, "  -h: Give this help message\n");
	fprintf(stderr, "  -q: Quiet\n");
	
	/***added from Ane code***/
	fprintf(stderr,"\n Covarion model options [default = none] :\n");
	fprintf(stderr, "  -v: #1 #2 = Tuffley & Steel/Huelsenbeck model with ON frequency and speed of covarion evolution.\n");
	fprintf(stderr, "      speed is the average number of switches per (average) substitution.\n\n");
	fprintf(stderr, "  -u: #1 #2 = Galtier model with proportion of covarion sites (the others are RAS) and switching rate nu\n");
	fprintf(stderr, "      nu is per expected substitution,\n");
	fprintf(stderr, "      but includes 'switches' back into the existing state\n");
	fprintf(stderr, "  -y: #1 #2 #3 #4 = General covarion model.  #1 is proportion of covarion sites (the others are RAS).\n"); 
	fprintf(stderr, "      #2 is switching rate from OFF to ON per expected substitution,\n");
	fprintf(stderr, "      #3 is switching rate from ON to OFF per expected substitution,\n");
	fprintf(stderr, "      #4 is switching rate among ON states per expected substitution,\n");
	fprintf(stderr, "      but includes 'switches' back into the existing state\n\n");

	/***added C20, C60 and SSF ***/
	fprintf(stderr,"\n Site-specific frequency (SSF) options [default = none] :\n");
	fprintf(stderr, "  -C20: #1 = number of sites to be simulated for each of the C20+F components;\n");
	fprintf(stderr, "      the alignment length specified by -l must be equal to this number multipled by 21.\n");
	fprintf(stderr, "      F is the amino acid frequency vector of the model specified by -m;\n");
	fprintf(stderr, "      Must add -p 21 to specify 21 partitions for the C20 model\n\n");
	fprintf(stderr, "  -C60: #1 = number of sites to be simulated for each of the C60+F components;\n");
	fprintf(stderr, "      the alignment length specified by -l must be equal to this number multipled by 61.\n");
	fprintf(stderr, "      F is the amino acid frequency vector of the model specified by -m;\n");
	fprintf(stderr, "      Must add -p 61 to specify 61 partitions for the C60 model\n\n");
	fprintf(stderr, "  -SSF: NAME = file name of the target amino acid site frequencies\n");
	fprintf(stderr, "      the alignment length specified by -l must be equal to the number of the lines in the site frequency file.\n");
	fprintf(stderr, "      Must add -p #1 where #1 must be equal to the alignment length specified by -l \n\n");
	fprintf(stderr, "  -wt: the weight file for C20+F or C60+F [default = none] :\n");
	

	/***************************/
	
	fprintf(stderr, "\n  treefile: name of tree file [default = trees on stdin]\n\n");
}


int whichAA(char c) {
  int i;
  const char AAset[] = {'A','C','D','E','F','G','H','I','K','L','M','N','P','Q','R','S','T','V','W','Y'};

  for (i=0; i<20; i++)
    if (c == AAset[i]) return i;

  return (-1);
}


void ReadParams(int argc, char**argv)
{
	int i, j;
	char ch, *P, st[4];
	double cov[4];/*added from Ane code*/

	model=NONE;

	scaleTrees=0;
	treeScale=0.0;
	scaleBranches=0;
	branchScale=0.0;
	
	maxPartitions=1;
	numPartitions=1;
	
	userSeed = 0;
	
	numCats=1;
	rateHetero=NoRates;
	catRate[0]=1.0;
	gammaShape=1.0;
	
	invariableSites=0;
	proportionInvariable = 0.0;
		
	/*added from Ane code*/
	covarion=NoCovarion;
	covarionSpeed=0.0;
	ONfreq=0.5;
	pcov=0.0;
	s01=0.0;
	s10=0.0;
	s11=0.0;
	/*******************/
	
	equalFreqs = 1;
	equalTstv = 1;
	tstv=0.50002;
	
	for (i = 0; i < NUM_AA_REL_RATES; i++) {
		aaRelativeRate[i] = 1.0;
	}
	
	for (i = 0; i < NUM_AA; i++) {
		aaFreq[i] = 1.0;
	}
	
	aaFreqSet = 0;
		
	numSites=-1;
	numDatasets=1;
	
	ancestorSeq=0;
	
	writeAncestors=0;
	writeRates=0;
	
	/*added from Ane code*/
	writeLengths=0;
	writeHidStates=0;	
	/*************************/
 	verbose=1;
	fileFormat = PHYLIPFormat;
	quiet=0;

	treeFile=0;
	textFile=0;
	SSF = 0;
	C20 = 0;
	C60 = 0;
	havingWeight=0;
	user_defined_aafreq = 0;
	int flag=0;

	for (i=1; i<argc; i++) {
		P=argv[i];
		if (*P=='-') {
			P++;
			ch=toupper(*P);
			P++;
			switch (ch) {
				case 'H':
					PrintTitle();
					PrintUsage();
					exit(0);
				break;
			case 'M':   /* use -mWAG or -mLG  Note: there is no space b/w -m and the rate models */
					if (GetStrParam(argc, argv, &i, P, st, 3)) {
						fprintf(stderr, "Bad (or missing) Model Code: %s\n\n", argv[i]);
						exit(0);
					}
					
					P=st;
					while (*P) {
						*P=toupper(*P);
						P++;
					}
					P=st;
		
					model=-1;
					for (j=F84; j<numModels; j++) {
						if (strncmp(P, modelNames[j], 3)==0) {
							model=j;
							if (model <= GTR) {
								isNucModel = 1;
								numStates = 4;
							} else {
								isNucModel = 0;
								numStates = 20;
							}
						} else if (strncmp(P, "REV", 3)==0) {
							model=GTR;
							isNucModel = 1;
							numStates = 4;
						}

					}
					if (model==-1) {
						fprintf(stderr, "Unknown Model: %s\n\n", argv[i]);
						exit(0);
					}

				break;
			}
		}
	}

	if (model==NONE) {
		fprintf(stderr, "No model has been specified (use the -m option)\n\n");
		PrintUsage();
		exit(0);
	}

	for (i=1; i<argc; i++) {
		P=argv[i];
		if (*P!='-') {
			if (treeFile) {
				fprintf(stderr, "Illegal command parameter: %s\n\n", argv[i]);
				PrintUsage();
				exit(0);
			}
			treeFile=1;
			strcpy(treeFileName, argv[i]);
		} else if (*P=='-' && toupper(*(P+1))=='X') {
			P++; P++;
			if (*P=='\0') {
				i++;
				P = argv[i];
			}
			textFile=1;
			strcpy(textFileName, P);
		} else if (*P=='-' && toupper(*(P+1))=='S' && toupper(*(P+2))=='S' && toupper(*(P+3))=='F') {
		  P++; P++; P++; P++;
		  if (*P=='\0') {
		    i++;
		    P = argv[i];
		  }
		  SSF = 1;
		  simuPartitionLen = 1; /* for per-site ssf partition length = 1 */
		  strcpy(site_aafreqFileName, P);
		} else if (*P=='-' && toupper(*(P+1))=='C' && (*(P+2))=='2' && (*(P+3))=='0') {
		  P++; P++; P++; P++;
		  C20 = 1;
		  if (GetIntParams(argc, argv, &i, P, 1, &simuPartitionLen) || simuPartitionLen<2) {
		    fprintf(stderr, "Each C20 component should have at least 1 site: %s\n\n", argv[i]);
		    exit(0);
		  }

		} else if (*P=='-' && toupper(*(P+1))=='C' && (*(P+2))=='6' && (*(P+3))=='0') {
		  P++; P++; P++; P++;
		  C60 = 1;
		  if (GetIntParams(argc, argv, &i, P, 1, &simuPartitionLen) || simuPartitionLen<2) {
		    fprintf(stderr, "Each C60 component should have at least 1 site: %s\n\n", argv[i]);
		    exit(0);
		  }
		}
		else if (*P=='-' && toupper(*(P+1))=='W' && toupper(*(P+2))=='T' ) {
		  P++; P++; P++;
		  if (*P=='\0') {
		    i++;
		    P = argv[i];
		  }
		  strcpy(CFweightFileName, P);
		  havingWeight = 1;
		}
		else if (*P=='-' && toupper(*(P+1))=='A' && toupper(*(P+2))=='A' && toupper(*(P+3))=='F' && toupper(*(P+4))=='R' && toupper(*(P+5))=='E' && toupper(*(P+6))=='Q') {
		  P++; P++; P++; P++; P++; P++; P++;
		  if (*P=='\0') {
		    i++;
		    P = argv[i];
		  }
		  strcpy(aaFreqFileName, P);
		  user_defined_aafreq = 1;
		}
  
		else {
			P++;
			ch=toupper(*P);
			P++;
			switch (ch) {
				case 'H':
				  /* already dealt with*/
				break;
				case 'M':
				  /*already dealt with*/
				break;
				case 'L':
					if (GetIntParams(argc, argv, &i, P, 1, &numSites) || numSites<1) {
						fprintf(stderr, "Bad (or missing) sequence length: %s\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'N':
					if (GetIntParams(argc, argv, &i, P, 1, &numDatasets) || numDatasets<1) {
						fprintf(stderr, "Bad (or missing) number of datasets: %s\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'P':
					if (GetIntParams(argc, argv, &i, P, 1, &maxPartitions) || maxPartitions < 1) {
						fprintf(stderr, "Bad number of partitions: %s\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'C':
					if (!isNucModel) {
						fprintf(stderr, "You can only have codon rates when using nucleotide models\n\n");
						exit(0);
					}
					if (rateHetero==GammaRates) {
						fprintf(stderr, "You can only have codon rates or gamma rates not both\n\n");
						exit(0);
					}
					numCats=3;
					rateHetero=CodonRates;
					if (GetDoubleParams(argc, argv, &i, P, 3, catRate) ||
						catRate[0] <= 0.0 || catRate[1] <= 0.0 || catRate[2] <= 0.0 ) {
						fprintf(stderr, "Bad Category Rates: %s\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'I':
					if (GetDoubleParams(argc, argv, &i, P, 1, &proportionInvariable) || 
							proportionInvariable < 0.0 || proportionInvariable >= 1.0) {
						fprintf(stderr, "Bad Proportion of Invariable Sites: %s\n\n", argv[i]);
						exit(0);
					}
					invariableSites = 1;
				break;
				case 'A':
				  if (flag==1) break;  /* to avoid conflict with -aafreq */
				  else {
					if (rateHetero==CodonRates) {
						fprintf(stderr, "You can only have codon rates or gamma rates not both\n\n");
						exit(0);
					}
					
					if (rateHetero==NoRates)
						rateHetero=GammaRates;
					if (GetDoubleParams(argc, argv, &i, P, 1, &gammaShape) || gammaShape<=0.0) {
						fprintf(stderr, "Bad Gamma Shape: %s\n\n", argv[i]);
						exit(0);
					}
					flag = 1;
				  }
				break;
				case 'G':
					if (rateHetero==CodonRates) {
						fprintf(stderr, "You can only have codon rates or gamma rates not both\n\n");
						exit(0);
					}
					
					rateHetero=DiscreteGammaRates;
					if (GetIntParams(argc, argv, &i, P, 1, &numCats) || numCats<2 || numCats>MAX_RATE_CATS) {
						fprintf(stderr, "Bad number of Gamma Categories: %s\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'F':
					if (isNucModel) {
						if (toupper(*P)=='E'){
							/* do nothing - equal freqs is default for nucleotides */
						} else {
							equalFreqs = 0;
							if (GetDoubleParams(argc, argv, &i, P, NUM_NUC, nucFreq)) {
								fprintf(stderr, "Bad Nucleotide Frequencies: %s\n\n", argv[i]);
								exit(0);
							}
						}
					} else {
						aaFreqSet = 1;
						if (toupper(*P)=='E'){      /* -fE */
							equalFreqs = 1;
							for(j=0;j<NUM_AA;j++) {
								aaFreq[j]=0.05;
							}
						} else {                    /* enter the freqs on the command line to set the freq */
							equalFreqs = 0;
							if (GetDoubleParams(argc, argv, &i, P, NUM_AA, aaFreq)) {
								fprintf(stderr, "Bad Amino Acid Frequencies: %s\n\n", argv[i]);
								exit(0);
							}
						}
					}
				break;
				case 'T':
					if (model != HKY && model != F84) {
						fprintf(stderr, "You can only have a transition/transversion ratio when using HKY or F84 models\n\n");
						exit(0);
					}
					equalTstv = 0;
					if (GetDoubleParams(argc, argv, &i, P, 1, &tstv)) {
						fprintf(stderr, "Bad Transition-Transversion Ratio: %s\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'R':
					if (model == GTR) {
						if (GetDoubleParams(argc, argv, &i, P, NUM_NUC_REL_RATES, nucRelativeRates)) {
							fprintf(stderr, "Bad General Nucleotide Rate Matrix: %s\n\n", argv[i]);
							exit(0);
						}
						if (nucRelativeRates[NUM_NUC_REL_RATES - 1]!=1.0) {
							for (j=0; j < NUM_NUC_REL_RATES - 1; j++) 
								nucRelativeRates[j] /= nucRelativeRates[NUM_NUC_REL_RATES - 1];
							nucRelativeRates[NUM_NUC_REL_RATES - 1] = 1.0;
						}
					} else if ( model == GENERAL) {
						if (GetDoubleParams(argc, argv, &i, P, NUM_AA_REL_RATES, aaRelativeRate)) {
							fprintf(stderr, "Bad General Amino Acid Rate Matrix: %s\n\n", argv[i]);
							exit(0);
						}
					} else {
						fprintf(stderr, "You can only have a general rate matrix when using GTR or GENERAL models\n\n");
						exit(0);
					}
				break;
				case 'D':
					scaleTrees=1;
					if (GetDoubleParams(argc, argv, &i, P, 1, &treeScale) || treeScale<=0.0) {
						fprintf(stderr, "Bad Total Tree Scale: %s\n\n", argv[i]);
						exit(0);
					}
					if (scaleBranches) {
						fprintf(stderr, "You can't specify both the -d and -s options\n\n");
						exit(0);
					}
				break;
				case 'S':
					scaleBranches=1;
					if (GetDoubleParams(argc, argv, &i, P, 1, &branchScale) || branchScale<=0.0) {
						fprintf(stderr, "Bad Branch Length Scale: %s\n\n", argv[i]);
						exit(0);
					}
					if (scaleTrees) {
						fprintf(stderr, "You can't specify both the -d and -s options\n\n");
						exit(0);
					}
				break;
				case 'K':
					if (GetIntParams(argc, argv, &i, P, 1, &ancestorSeq) || ancestorSeq<1) {
						fprintf(stderr, "Bad ancestral sequence number: %s\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'Z':
					userSeed = 1;
					if (GetUnsignedLongParams(argc, argv, &i, P, 1, &randomSeed)) {
						fprintf(stderr, "Bad random number generator seed: %s\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'O':
					switch (toupper(*P)) {
						case 'P': fileFormat=PHYLIPFormat; break;
						case 'R': fileFormat=RelaxedFormat; break;
						case 'N': fileFormat=NEXUSFormat; break;
						default:					
							fprintf(stderr, "Unknown output format: %s\n\n", argv[i]);
							PrintUsage();
							exit(0);
					}
				break;

			case 'W':
				  switch (toupper(*P)) {
				  case 'A': writeAncestors=1; break;
				  case 'R': writeRates=1; break;
				    
				    // added from Ane code 
				  case 'H': writeHidStates=1; break;
				  case 'L': writeLengths=1; break;

				    
				  default:					
				    fprintf(stderr, "Unknown write mode: %s\n\n", argv[i]);
				    PrintUsage();
				    exit(0);
				  }
				break;
				
			case 'Q':
			  quiet=1;
			  break;
				
				/*added from Ane code*/
				case 'V':/*Huelsenbeck covarion model*/
				  if (GetDoubleParams(argc, argv, &i, P, 2, cov) ||cov[0]<=0.0 || cov[0]>=1.0||cov[1]<= 0.0) {
				    fprintf(stderr, "Bad Covarion Parameters: %s %s \n\n", argv[i-1], argv[i]);
				    exit(0);
				  }
					covarion=Huel;
					ONfreq=cov[0];
					covarionSpeed=cov[1];
				break;
				/***************************/
				case 'U':/*Galtier covarion model*/
				  if (GetDoubleParams(argc, argv, &i, P, 2, cov) ||cov[0]<0.0 || cov[0]>1.0||cov[1]<= 0.0) {
				    fprintf(stderr, "Bad Covarion Parameters: %s %s \n\n", argv[i-1], argv[i]);
				    exit(0);
				  }
					covarion=Galt;
					pcov=cov[0];
					covarionSpeed=cov[1];
				break;
				case 'Y':/*General covarion model*/
				  if (GetDoubleParams(argc, argv, &i, P, 4, cov) ||cov[0]<0.0 || cov[0]>1.0||cov[1]< 0.0||cov[2]< 0.0||cov[3]< 0.0) {
				    fprintf(stderr, "Bad Covarion Parameters: %s %s %s %s\n\n", argv[i-3],argv[i-2],argv[i-1], argv[i]);
				    exit(0);
				  }
					covarion=General;
					pcov=cov[0];
					s01=cov[1];
					s10=cov[2];
					s11=cov[3];
				break;
				
				default:
					fprintf(stderr, "Illegal command parameter: %s\n\n", argv[i]);
					PrintUsage();
					exit(0);
				break;
			}
		}
	}
	if ((covarion==Galt)|(covarion==General)) {/*checking we have discrete gamma, independent of parameter order*/
		if (rateHetero != DiscreteGammaRates) {
			fprintf(stderr, "You must specify discrete gamma rates for the Galtier or General covarion model\n\n");
			exit(0);
		}
		rateHetero=NoRates;/*turn off the RAS stuff when sequences evolve: we probably don't want this on top of the Galtier/General model*/
	}
}

void PrintVerbose(FILE *fv)
{
	int i;
	
	if (numStates == 4)
		fprintf(fv, "Simulations of %d taxa, %d nucleotides\n", numTaxa, numSites);
	else 
		fprintf(fv, "Simulations of %d taxa, %d amino acids\n", numTaxa, numSites);
	fprintf(fv, "  for %d tree(s) with %d dataset(s) per tree\n", numTrees, numDatasets);
	/*
	if (numPartitions > 1) {
		fprintf(fv, "  and %d partitions (and trees) per dataset\n", numPartitions);
		fprintf(fv, "    Partition  No. Sites  Relative Rate\n");
		for (i = 0; i < numPartitions; i++) 
			fprintf(fv, "    %4d       %7d    %f\n", i+1, partitionLengths[i], partitionRates[i]);
	}
	*/
	fputc('\n', fv);
	
	
	if (scaleTrees) {
		fprintf(fv, "Branch lengths of trees scaled so that tree is %G from root to tip\n\n", treeScale);
	} else if (scaleBranches) {
		fprintf(fv, "Branch lengths of trees multiplied by %G\n\n", branchScale);
	} else {
		fprintf(fv, "Branch lengths assumed to be number of substitutions per site\n\n");
	}
	if (rateHetero==CodonRates) {
		fprintf(fv, "Codon position rate heterogeneity:\n");
		fprintf(fv, "    rates = 1:%f 2:%f 3:%f\n", catRate[0], catRate[1], catRate[2]);
	} else if (rateHetero==GammaRates) {
		fprintf(fv, "Continuous gamma rate heterogeneity:\n");
		fprintf(fv, "    shape = %f\n", gammaShape);
	} else if (rateHetero==DiscreteGammaRates) {
		fprintf(fv, "Discrete gamma rate heterogeneity:\n");
		fprintf(fv, "    shape = %f, %d categories\n", gammaShape, numCats);
	} else
		fprintf(fv, "Rate homogeneity of sites.\n");
	if (invariableSites) {
		fprintf(fv, "Invariable sites model:\n");
		fprintf(fv, "    proportion invariable = %f\n", proportionInvariable);
	}
	fprintf(fv, "Model = %s\n", modelTitles[model]);
	if (isNucModel) {
		if (equalTstv) {
			fprintf(fv, "  Rate of transitions and transversions equal:\n");
		}
		if (model==F84) {
			fprintf(fv, "  transition/transversion ratio = %G (K=%G)\n", tstv, kappa);
		} else if (model==HKY) {
			fprintf(fv, "  transition/transversion ratio = %G (kappa=%G)\n", tstv, kappa);
		} else if (model==GTR) {
			fprintf(fv, "  rate matrix = gamma1:%7.4f alpha1:%7.4f  beta1:%7.4f\n", nucRelativeRates[0], nucRelativeRates[1], nucRelativeRates[2]);
			fprintf(fv, "                                beta2:%7.4f alpha2:%7.4f\n", nucRelativeRates[3], nucRelativeRates[4]);
			fprintf(fv, "                                              gamma2: %7.4f\n", nucRelativeRates[5]);
		}

		if (equalFreqs) {
			fprintf(fv, "  with nucleotide frequencies equal.\n");
		} else {
			fprintf(fv, "  with nucleotide frequencies specified as:\n");
			fprintf(fv, "  A=%G C=%G G=%G T=%G\n\n", freq[A], freq[C], freq[G], freq[T]);
		}
	} else {
		if (aaFreqSet) {
			if (equalFreqs) {
				fprintf(fv, "  with amino acid frequencies equal.\n\n");
			} else {
				fprintf(fv, "  with amino acid frequencies specified as:\n");
				fprintf(fv, "  ");
				for (i = 0; i < NUM_AA; i++) {
					fprintf(fv, " %c=%G", aminoAcids[i], freq[i]);
				}
				fprintf(fv, "\n\n");
			}
		}
	}
	switch (covarion) {
		case Huel:
			fprintf(fv, "Tuffley & Steel/Huelsenbeck covarion evolution:\n  ON frequency = %G\n",ONfreq);
			fprintf(fv, "  speed (average number of switches / substitution) = %G\n\n",covarionSpeed);
		break;
		case Galt:
			fprintf(fv, "Galtier covarion evolution:\n  proportion covarion sites = %G\n",pcov);
			fprintf(fv, "  speed (average number of switches / substitution) = %G\n",covarionSpeed);
			fprintf(fv, "Discrete gamma parameters for Galtier model:\n");
			fprintf(fv, "    shape = %f, %d categories\n\n", gammaShape, numCats);
		break;
		case General:
			fprintf(fv, "General covarion evolution:\n  proportion covarion sites = %G\n",pcov);
			fprintf(fv, "  s01 = %G, s10 = %G, s11 = %G\n",s01,s10,s11);
			fprintf(fv, "Discrete gamma parameters for General model:\n");
			fprintf(fv, "    shape = %f, %d categories\n\n", gammaShape, numCats);
		break;
		case NoCovarion:
			fprintf(fv, "No covarion evolution \n\n");
		break;
	}
}

void ReadFileParams()
{
	char ch, st[256];
	
	hasAlignment=0;
	
	ch=fgetc(stdin);
	while (!feof(stdin) && isspace(ch)) 
		ch=fgetc(stdin);
		
	ungetc(ch, stdin);

	if (ch!='(' && isdigit(ch)) {
		fgets(st, 255, stdin);
		if ( sscanf( st, " %d %d", &numSequences, &numAlignmentSites)!=2 ) {
			fprintf(stderr, "Unable to read parameters from standard input\n");
			exit(0);
		}
		hasAlignment=1;
	}
		
}

void AllocateMemory()
{
	int i;
	
	names=(char **)AllocMem(sizeof(char *)*numSequences, "names", "AllocateMemory", 0);
	sequences=(char **)AllocMem(sizeof(char *)*numSequences, "sequences", "AllocateMemory", 0);
	for (i=0; i<numSequences; i++) {
		names[i]=(char *)AllocMem(sizeof(char)*(MAX_NAME_LEN+1), "names[]", "AllocateMemory", 0);
		sequences[i]=(char *)AllocMem(sizeof(char)*numAlignmentSites, "sequences[]", "AllocateMemory", 0);
	}
}


void ReadFile()
{
	int n, b, i;
	char ch;
		
	n=0;
	do {
		ch=fgetc(stdin);
		while ( !feof(stdin) && isspace(ch)) 
			ch=fgetc(stdin);
			
		if ( feof(stdin) ) {
			fprintf(stderr, "Unexpected end of file on standard input\n"); 
			exit(0);
		}
	
		i=0;
		while ( i<MAX_NAME_LEN && !feof(stdin) && !isspace(ch) ) {
			names[n][i]=ch;
			ch=fgetc(stdin);
			i++;
		}
		names[n][i]='\0';
		if (i==0) {
			fprintf(stderr, "Name missing for species %d\n", n+1);
			exit(0);
		}
		while (!feof(stdin) && isspace(ch))
			ch=fgetc(stdin);
		
		if ( feof(stdin) ) {
			fprintf(stderr, "Unexpected end of file on standard input\n");
			exit(0);
		}
		
		b=0;
		while ( !feof(stdin) && b<numAlignmentSites) {
			if ( !isspace(ch) ) {
				sequences[n][b]=ch;
				b++;
			}
			ch=toupper(fgetc(stdin));
		}
		
		if ( b<numAlignmentSites ) {
			fprintf(stderr, "Unexpected end of file on standard input\n");
			exit(0);
		}
		
		/*fprintf(stderr, "%d: %s, bases read: %d\n", n+1, names[n], b);*/
		n++;
		
		if ( n<numSequences && feof(stdin) ) {
			fprintf(stderr, "Too few sequences in input file\n");
			exit(0);
		}
	} while ( n<numSequences );
}

int OpenTreeFile()
{
	char st[256];
	int n;
		
	if (treeFile) {
		if ( (tree_fv=fopen(treeFileName, "rt"))==NULL ) {
			fprintf(stderr, "Error opening tree file: '%s'\n", treeFileName);
			exit(0);
		}
		n=CountTrees(tree_fv);
	} else {
		tree_fv=stdin;
		if (hasAlignment) {
			fgets(st, 255, stdin);
			if ( sscanf(st, " %d ", &n)!=1 ) {
				fprintf(stderr, "Tree is missing from end of sequence file\n");
				exit(0);
			}
		} else
			n=CountTrees(stdin);
	}
	
	return n;
}


int sitefreqlines (char *site_aafreqFileName)
{
  int count=0;
  FILE *infile;
  char buf[5000];

  infile = fopen (site_aafreqFileName, "r");   

  while (fgets(buf,5000,infile)!=NULL) 
    count++;

  printf("count=%d\n", count);
  fclose (infile);

  return (count);
}


int main(int argc, char **argv)
{
	int i, j, k, treeNo, sumLength;
	char ch;
	TTree **treeSet;
	FILE *text_fv;
	clock_t totalStart;
	double totalSecs, scale, sum;
	char *ancestor;

	int how_many_sites;

	FILE *wtfile;
	char buf[5000];
	char *pch; 
	float weightC20F[21], weightC60F[61], weightsum=0.;

	totalStart = clock();

	ReadParams(argc, argv);

	if (C20 == 1) {
	  aaFreqSet = 1;
	  if (maxPartitions != 21) {
	    printf("\nError: the number of partitions in C20+F should be 21. Use -p 21 on the command line \n\n");
	    exit(0);
	  }

	  if (havingWeight != 1) 
	    if (simuPartitionLen * 21 != numSites) {
	      printf("\nError: the sum of the 21 partition lengths in C20+F is not equal to the number of sites specified.\n\n");
	      exit(0);
	    }
	  
	  if (havingWeight == 1) {
	    if ((wtfile = fopen(CFweightFileName,"r"))==NULL) {
	      printf("\nError: mixture weight file does not exist.\n");
	      exit(0);
	    }
	
	    fgets(buf, 5000, wtfile);
	    pch = strtok(buf, " ");
	    
	    i = 0;
	    while (pch != NULL){
	      weightC20F[i++] = atof(pch);
	      weightsum += atof(pch);
	      pch = strtok (NULL, " ");
	    }
	    
	    if (i > 22 || i < 21 || weightsum > 1.05 || weightsum < 0.95) {   /* to accomodate some weight file has an extra space at the end of the file */
	      printf("%d\t%.4f\n", i, weightsum);
	      printf("\nError: mixture weight file incorrect.\n");
	      exit(0);
	    }
	    
	  }
	  
	}

	if (C60 == 1) {
	  aaFreqSet = 1;
	  if (maxPartitions != 61) {
	    printf("\nError: the number of partitions in C60+F should be 61. Use -p 61 on the command line \n\n");
	    exit(0);
	  }
	  
	  if (havingWeight != 1) 
	    if (simuPartitionLen * 61 != numSites) {
	      printf("\nError: the sum of the 61 partition lengths in C60+F is not equal to the number of sites specified\n\n");
	      exit(0);
	    }

	  if (havingWeight == 1) {
	    if ((wtfile = fopen(CFweightFileName,"r"))==NULL) {
	      printf("\nError: mixture weight file does not exist.\n");
	      exit(0);
	    }
	
	    fgets(buf, 5000, wtfile);
	    pch = strtok(buf, " ");
	    
	    i = 0;
	    while (pch != NULL){
	      weightC60F[i++] += atof(pch);
	      weightsum += atof(pch);
	      pch = strtok (NULL, " ");
	    }
	    
	    if (i > 62 || i < 61 || weightsum > 1.05 || weightsum < 0.95) {  /* to accomodate some weight file has an extra space at the end of the file */
	      printf("\nError: mixture weight file incorrect.\n");
	      exit(0);
	    }
	    
	  }
	}

	if (SSF == 1) {
	  how_many_sites = sitefreqlines (site_aafreqFileName);
	  aaFreqSet = 1;
	  if (maxPartitions != numSites) {
	    printf("\nError: the number of partitions in SSF for per-site partition should be equal to the number of sites specified. -p and -l should list the same number \n\n");
	    exit(0);
	  }
	  if (maxPartitions != how_many_sites) {
	    printf("\nError: the number of partitions in SSF for per-site partition should be equal to the number of site-frequency lines in the %s file.\nThe number of site-frequency lines in the %s file is %d, so use the same number in both -p and -l.\n\n", site_aafreqFileName, site_aafreqFileName, how_many_sites);
	    exit(0);
	  }
	}

	if ( aaFreqSet == 1) 
	  for(j=0;j<NUM_AA;j++)
	    aaFreq[j]=0.05;   /* the initial aafreq will be ignored */


	if (rateHetero == CodonRates && invariableSites) {
		fprintf(stderr, "Invariable sites model cannot be used with codon rate heterogeneity.\n");
		exit(0);
	}

	if (writeAncestors && fileFormat == NEXUSFormat) {
		fprintf(stderr, "Warning - When writing ancestral sequences, relaxed PHYLIP format is used.\n");
	}

	if (writeAncestors && maxPartitions > 1) {
		fprintf(stderr, "Writing ancestral sequences can only be used for a single partition.\n");
		exit(0);
	}
			
	if (!userSeed)
		randomSeed = CreateSeed();
		
	SetSeed(randomSeed);

	if (!quiet)
 		PrintTitle();
	
	if (!treeFile) {
		ReadFileParams();
	}
	
	if ((ancestorSeq>0 && !hasAlignment) || ancestorSeq>numSequences) {
		fprintf(stderr, "Bad ancestral sequence number\n");
		exit(0);
	}
	
	if (textFile) {
		if ( (text_fv=fopen(textFileName, "rt"))==NULL ) {
			fprintf(stderr, "Error opening text file for insertion into output: '%s'\n", textFileName);
			exit(0);
		}
	}

	ancestor=NULL;
	if (hasAlignment) {
		AllocateMemory();	
		ReadFile();
		
		if (numSites<0)
			numSites=numAlignmentSites;		
			
		if (ancestorSeq>0) {
			if (numSites!=numAlignmentSites) {
				fprintf(stderr, "Ancestral sequence is of a different length to the simulated sequences\n");
				exit(0);
			}
			ancestor=sequences[ancestorSeq-1];
		}
	} else if (numSites<0)
		numSites=1000;
	

	SetModel(model);   /* for site partition models this function is ignored */

	numTaxa=-1;
	scale=1.0;
	
	treeSet = (TTree **)malloc(sizeof(TTree **) * maxPartitions);
	if (treeSet==NULL) {
		fprintf(stderr, "Out of memory\n");
		exit(0);
	}
	
	partitionLengths = (int *)malloc(sizeof(int) * maxPartitions);
	if (partitionLengths==NULL) {
		fprintf(stderr, "Out of memory\n");
		exit(0);
	}
	
	partitionRates = (double *)malloc(sizeof(double) * maxPartitions);
	if (partitionRates==NULL) {
		fprintf(stderr, "Out of memory\n");
		exit(0);
	}
	
	for (i = 0; i < maxPartitions; i++) {
		if ((treeSet[i]=NewTree())==NULL) {
			fprintf(stderr, "Out of memory\n");
			exit(0);
		}
	}
			
	numTrees = OpenTreeFile();

	CreateRates();
	
	treeNo=0;
	do {
		partitionLengths[0] = -1;
		ReadTree(tree_fv, treeSet[0], treeNo+1, 0, NULL, &partitionLengths[0], &partitionRates[0]);
		
		partitionLengths[0] = simuPartitionLen;
		
		if (havingWeight == 1 && C20 == 1) 
		  partitionLengths[0] = numSites * weightC20F[0]; /* in IQtree output, weights[0] is the wt for LG_F, weights[1..60] is the wt for the C20 components; */

		if (havingWeight == 1 && C60 == 1) 
		  partitionLengths[0] = numSites * weightC60F[0]; /* in IQtree output, weights[0] is the wt for LG_F, weights[1..60] is the wt for the C60 components; */

		if (treeNo==0) {
			numTaxa=treeSet[0]->numTips;
			
			if (!quiet)
				fprintf(stderr, "Random number generator seed: %ld\n\n", randomSeed);
				
			if (fileFormat == NEXUSFormat) {
				fprintf(stdout, "#NEXUS\n");
				fprintf(stdout, "[\nGenerated by %s %s\n\n", PROGRAM_NAME, VERSION_NUMBER);
				PrintVerbose(stdout);
				fprintf(stdout, "]\n\n");
			}
		} else if (treeSet[0]->numTips != numTaxa) {
			fprintf(stderr, "All trees must have the same number of tips.\n");
			exit(0);
		}
		
		if (maxPartitions == 1) {
			if (partitionLengths[0] != -1) {
				fprintf(stderr, "\nWARNING: The treefile contained partion lengths but only one partition\n");
				fprintf(stderr, "was specified.\n");
			}
			partitionLengths[0] = numSites;
		}

		sumLength = partitionLengths[0];
		i = 1;

		while (sumLength < numSites && i <= maxPartitions) {

		  rewind(tree_fv);
		  if (!IsTreeAvail(tree_fv)) {
		    fprintf(stderr, "\nA set of trees number %d had less partition length (%d) than\n", treeNo + 1, sumLength);
		    fprintf(stderr, "was required to make a sequence of length %d.\n", numSites);
		    exit(0);
		  }
				
		  ReadTree(tree_fv, treeSet[i], treeNo+1, treeSet[0]->numTips, treeSet[0]->names, 
						&partitionLengths[i], &partitionRates[i]);
			
		  if (treeSet[i]->numTips != numTaxa) {
		    fprintf(stderr, "All trees must have the same number of tips.\n");
		    exit(0);
		  }
			
		  partitionLengths[i] = simuPartitionLen;  /* partition length = 1 for site partition model */

		  if (havingWeight == 1 && C20 == 1) {
		    if (i < maxPartitions-1) partitionLengths[i] = numSites * weightC20F[i]; 
		    if (i == maxPartitions-1) partitionLengths[i] = numSites - sumLength;
		  }
		  if (havingWeight == 1 && C60 == 1) {
		    if (i < maxPartitions-1) partitionLengths[i] = numSites * weightC60F[i]; 
		    if (i == maxPartitions-1) partitionLengths[i] = numSites - sumLength;
		  }

		  sumLength += partitionLengths[i];
		  i++;
		}

		if (i > maxPartitions) {
			fprintf(stderr, "\nA set of trees number %d had more partitions (%d) than\n", treeNo + 1, i);
			fprintf(stderr, "was specified in the user options (%d).\n", maxPartitions);
		}
		numPartitions = i;
				
		if (sumLength != numSites) {
			fprintf(stderr, "The sum of the partition lengths in the treefile does not equal\n");
			fprintf(stderr, "the specified number of sites.\n");
			exit(0);
		}
			
		for (i = 0; i < numPartitions; i++) {
		  CreateSequences(treeSet[i], partitionLengths[i]);   /* use 1 set tree for per-site partition model */ 
			
			/*added from Ane code*/
			if (covarion) {
			  CreateHiddenStates(treeSet[i], partitionLengths[i]);
			}
			/***********************/
		}
		if (numPartitions > 1) {
			sum = 0.0;
			for (i = 0; i < numPartitions; i++)
				sum += partitionRates[i] * partitionLengths[i];
				
			for (i = 0; i < numPartitions; i++)
				partitionRates[i] *= numSites / sum;
		}
		
		if (treeNo==0 && verbose && !quiet) {
		  	PrintVerbose(stderr);
			InitProgressBar(numTrees*numDatasets);
			DrawProgressBar();
		}

		for (i=0; i<numDatasets; i++) {
			SetCategories();
			
			k = 0;
			for (j = 0; j < numPartitions; j++) {

			  /* if (j > 0)*/ SetModel1(model, j); /* let each partition takes the freq from the given freq for the partition */
			  
			  scale = partitionRates[j];

			  if (scaleTrees) { 
			    if (!treeSet[j]->rooted) {
			      fprintf(stderr, "To scale tree length, they must be rooted and ultrametric.\n");
			      exit(0);
			    }
			    scale *= treeScale/treeSet[0]->totalLength;
			  } else if (scaleBranches)
			    scale *= branchScale;
				
				/*aded from Ane code*/
				if (covarion) {
				  covarionRate = scale*covarionSpeed;
				  EvolveHiddenStates(treeSet[j], k, partitionLengths[j]);
     				}
				if (covarion==Huel) {
				  scale /= ONfreq;
				}
			       /***********************/

				EvolveSequences(treeSet[j], k, partitionLengths[j], scale, ancestor);
				k += partitionLengths[j];

				/*MS, 12/5/08*/
				if (writeHidStates && covarion) {
				  fprintf(stderr,"\nPartition %d hidden states:\n",j);
				  WriteHiddenStates(stderr, treeSet[j],partitionLengths[j],treeSet[0]);
				}
				if (writeLengths && covarion) {
				  fprintf(stderr,"\nPartition %d effective lengths:\n",j);
				  WriteAncestralEffectiveLengths(stderr, treeSet[j],partitionLengths[j],treeSet[0]);
				}/*MS, 12/5/08*/

			}
			
			if (writeAncestors)
				WriteAncestralSequences(stdout, treeSet[0]);
			else
				WriteSequences(stdout, (numTrees > 1 ? treeNo+1 : -1), (numDatasets > 1 ? i+1 : -1), treeSet, partitionLengths);

			if (writeRates) {
				WriteRates(stderr);
			}
			
			/*added from Ane code, but this doesn't work with more than one partition (12/5/08)*/
			/*if (writeHidStates && covarion) {
				WriteHiddenStates(stderr, treeSet[0]);
			}
			if (writeLengths && covarion) {
				WriteAncestralEffectiveLengths(stderr, treeSet[0]);
				}*/
			/*************************/
			
			if (textFile) {
				while (!feof(text_fv)) {
					ch = fgetc(text_fv);
					if (!feof(text_fv))
						fputc(ch, stdout);
				}
				fputc('\n', stdout);
				rewind(text_fv);
			}
			
			if (verbose && !quiet)
				ProgressBar();
		}
				
		/*for (i = 0; i < numPartitions; i++)
		  DisposeTree(treeSet[i]);*//*MS: removed 15/5/08*/
			
		treeNo++;
	}  while (IsTreeAvail(tree_fv));


	/*MS: added 15/5/08*/
	for (i = 0; i < maxPartitions; i++)
		FreeTree(treeSet[i]);
	free(treeSet);
	free(partitionLengths);
	free(partitionRates);
	FreeEvolve();
	/*****/

	if (treeFile)
		fclose(tree_fv);

	if (textFile)
		fclose(text_fv);

	totalSecs = (double)(clock() - totalStart) / CLOCKS_PER_SEC;
	if (!quiet) {
		fprintf(stderr, "Time taken: %G seconds\n", totalSecs);
		if (verboseMemory)
			fprintf(stderr, "Total memory used: %ld\n", totalMem);
	}
	
	return 0;
}
